#lang racket/base
;;; SPDX-License-Identifier: LGPL-3.0-or-later
;;; SPDX-FileCopyrightText: Copyright © 2021-2024 Tony Garnock-Jones <tonyg@leastfixedpoint.com>
;;; Smart constructors for compound patterns.

(provide rec
         arr
         dict
         discard
         bind

         rec*
         arr*
         dict*
         discard*
         bind*

         literal->literal-pattern
         literal-pattern->literal
         lit)

(require racket/dict)
(require racket/match)

(require preserves)
(require preserves-schema)

(require syndicate/schemas/dataspacePatterns)

(define (items->entries pats)
  (define npats (length pats))
  (define entries
    (for/fold [(entries (hash))] [(i (in-naturals)) (p (in-list pats))]
      (if (Pattern-discard? p) entries (hash-set entries i p))))
  (if (and (positive? npats) (not (hash-has-key? entries (- npats 1))))
      (hash-set entries (- npats 1) (Pattern-discard))
      entries))

(define (entries->items entries)
  (let/ec return
    (define keys (hash-keys entries))
    (for [(k (in-list keys))] (unless (exact-nonnegative-integer? k) (return (void))))
    (define max-key (apply max -1 keys))
    (for/list [(i (in-range 0 (+ max-key 1)))]
      (hash-ref entries i (lambda () (return void))))))

(define (rec label pats) (Pattern-group (GroupType-rec label) (items->entries pats)))
(define (arr pats) (Pattern-group (GroupType-arr) (items->entries pats)))
(define (dict pats) (Pattern-group (GroupType-dict) pats))
(define (discard) (Pattern-discard))
(define (bind [pat (discard)]) (Pattern-bind pat))

(define (rec* label pats) (rec 'group (list (rec 'rec (list (lit label))) (dict (items->entries pats)))))
(define (arr* pats) (rec 'group (list (rec 'arr (list)) (dict (items->entries pats)))))
(define (dict* pats) (rec 'group (list (rec 'dict (list)) (dict pats))))
(define (discard*) (rec '_ (list)))
(define (bind* [pat (discard)]) (rec 'bind (list pat)))

(define (literal->literal-pattern v)
  (let walk ((v (->preserve v)))
    (match v
      [(record label fs) (rec label (map walk fs))]
      [(? list? vs) (arr (map walk vs))]
      [(? dict? d) (dict (for/hash [((k v) (in-hash d))] (values k (walk v))))]
      [other (Pattern-lit (parse-AnyAtom! other))])))

(define lit literal->literal-pattern)

(define (literal-pattern->literal p)
  (let/ec return
    (define (e->i e)
      (match (entries->items e)
        [(? void?) (return (void))]
        [items items]))
    (let walk ((p p))
      (match p
        [(Pattern-discard) (return (void))]
        [(Pattern-bind pp) (walk pp)]
        [(Pattern-lit a) (->preserve a)]
        [(Pattern-group (GroupType-rec label) ps) (record label (map walk (e->i ps)))]
        [(Pattern-group (GroupType-arr) ps) (map walk (e->i ps))]
        [(Pattern-group (GroupType-dict) d) (for/hash [((k pp) (in-hash d))]
                                              (values k (walk pp)))]))))
